
/*  xmms-dxr2
 *
 *  Dxr2 DVD plugin for XMMS  (C) Jack Whitham, 2001
 *
 *
 *  Based on the cdaudio plugin
 *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "xmms-dxr2.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>
#include <xmms/util.h>
#include <player.h>
#include <ifo_parse.h>
#include <dxr2/interface.h>
#include <buffer.h>
#include <control.h>
#include <read_config.h>

#ifdef HAVE_LINUX_CDROM_H
#include <linux/cdrom.h>
#elif defined HAVE_SYS_CDIO_H
#include <sys/cdio.h>
#endif

#define TITLE_IDENTIFIER		"DVD Title "
#define VTS_DIRECT_IDENTIFIER	"DVD VTS "
#define VIDEO_TS_NAME			"/video_ts/video_ts.vob"
#define VTS_NAME				"/video_ts/vts_%02d_%d.vob" 
#define STATE_UNINITIALISED		0
#define STATE_CONFIG_ERROR		1
#define STATE_ERROR				2
#define STATE_READY				3
#define INFO_UNKNOWN			0
#define INFO_TITLE				1
#define INFO_VTS_DIRECT			2
#define INFO_VOB_DIRECT			3
#define DEFAULT_LENGTH			( 10 * 60 * 1000 )
#define DEFAULT_SUB_LENGTH		600


static void minidxr2_init(void);
static int is_our_file(char *filename);
static GList *scan_dir(char *dir);
static void play_file(char *filename);
static void stop(void);
static void minidxr2_pause(short p);
static void seek(int time);
static int get_time(void);
static void get_song_info(char *filename, char **title, int *length);
static void cleanup(void);
static int get_information ( char * filename , int * major , int * minor ) ;

InputPlugin minidxr2_ip =
{
	NULL,
	NULL,
	NULL, /* Description */
	minidxr2_init,
	NULL,				/* about */
	NULL , /* minidxr2_configure, */
	is_our_file,
	scan_dir,
	play_file,
	stop,
	minidxr2_pause,
	seek,
	NULL,				/* set_eq */
	get_time,
	NULL ,
	NULL ,
	cleanup,
	NULL,				/* obsolete */
	NULL,				/* add_vis_pcm */
	NULL,				/* set_info, filled in by xmms */
	NULL,				/* set_info_text, filled in by xmms */
	get_song_info,
	NULL, /*  minidxr2_fileinfo, */	/* file_info_box */
	NULL				/* output plugin handle */
};

/* CDDAConfig minidxr2_cfg; */

static pthread_t read_thread ;
static pthread_t write_thread ;
static int state = STATE_UNINITIALISED ;
static player_info_t * player_info = NULL ;
static char * config_file_error = NULL ;
static int current_audio_track = 1 ;
static int current_time_length = 0 ;
static int current_audio_freq = 48000 ;

InputPlugin * get_iplugin_info(void)
{
	minidxr2_ip . description = g_strdup_printf( "Dxr2 DVD Player %s", VERSION);
	return &minidxr2_ip;
}

static void minidxr2_deferred_init ( void )
{
	char				* err ;
	pthread_attr_t		thread_attrib ;

	if ( state == STATE_CONFIG_ERROR )
	{
		static GtkWidget * dialog;
		gchar			 * text ;

		state = STATE_ERROR ;

		if (dialog != NULL)
			return;

		if ( config_file_error == NULL )
		{
			config_file_error = "<unknown>" ;
		}
		text = g_strdup_printf ( "Error reading Dxr2 config file %s:\n"
						"  %s\n"
						"See Dxr2 documentation for an explanation of how\n"
						"to create this file.\n" , config_file_error , dxr2_config_file ) ;
		dialog = xmms_show_message ( "Error" ,
					text , "Ok" , FALSE , NULL , NULL ) ;
		g_free ( text ) ;

		gtk_signal_connect ( GTK_OBJECT ( dialog ) , "destroy" ,
				   GTK_SIGNAL_FUNC ( gtk_widget_destroyed ) ,
				   & dialog ) ;
		return ;
	}

	if ( state != STATE_UNINITIALISED )
	{
		return ;
	}

	buf_setup ( player_info ) ;
	ifo_setup ( player_info ) ;

	err = NULL ;
	player_info -> out_plugin = 
			output_setup ( (void *) player_info , (output_plugin_fn) dxr2_plugin , & err ) ;
	if ( err != NULL )
	{
		static GtkWidget * dialog;

		state = STATE_ERROR ;

		if (dialog != NULL)
			return;

		dialog = xmms_show_message ( "Error" ,
					"Error initialising Dxr2 output.\n" , "Ok" , FALSE , NULL , NULL ) ;

		gtk_signal_connect ( GTK_OBJECT ( dialog ) , "destroy" ,
				   GTK_SIGNAL_FUNC ( gtk_widget_destroyed ) ,
				   & dialog ) ;
		return ;
	}

	output_css_do_reset ( player_info -> out_plugin ,
					player_info -> dvd_device ) ;
	state = STATE_READY ;

	player_info -> h_state = DVD_UNLOADED ;

	/* Thread setup */
	pthread_attr_init(&thread_attrib);
    pthread_attr_setschedpolicy(&thread_attrib, SCHED_OTHER);
	pthread_create( &read_thread, &thread_attrib, buf_reader , (void*) player_info);
	pthread_create( &write_thread, &thread_attrib, buf_writer , (void*) player_info);
}


static void minidxr2_init(void)
{
	state = STATE_UNINITIALISED ;

	player_info = malloc ( sizeof ( player_info_t ) ) ;
	memset ( player_info , 0 , sizeof ( player_info_t ) ) ;

	/* Load in information that cannot be changed
	* (firmware file, mountpoint, etc) and defaults for audio settings */
	config_file_error = read_config ( player_info , dxr2_config_file ) ;
	if ( config_file_error != NULL )
	{
		config_file_error = read_config ( player_info , alt_config_file ) ;
		if ( config_file_error != NULL )
		{
			state = STATE_CONFIG_ERROR ;
		}
	}

/*	if ((cfgfile = xmms_cfg_open_default_file()))
	{
		xmms_cfg_read_string(cfgfile, "DXR2", "device", &minidxr2_cfg.device);
		xmms_cfg_read_string(cfgfile, "CDDA", "directory", &minidxr2_cfg.directory);
		xmms_cfg_read_boolean(cfgfile, "CDDA", "use_oss_mixer", &minidxr2_cfg.use_oss_mixer);
		xmms_cfg_read_string(cfgfile, "CDDA", "name_format", &minidxr2_cfg.name_format);
		xmms_cfg_read_boolean(cfgfile, "CDDA", "use_cddb", &minidxr2_cfg.use_cddb);
		xmms_cfg_read_string(cfgfile, "CDDA", "cddb_server", &minidxr2_cfg.cddb_server);
		xmms_cfg_read_int(cfgfile, "CDDA", "cddb_protocol_level", &minidxr2_cfg.cddb_protocol_level);
		xmms_cfg_read_string(cfgfile, "CDDA", "cdin_server", &minidxr2_cfg.cdin_server);
		xmms_cfg_free(cfgfile);
	} */
}

static void cleanup(void)
{
	if (( state != STATE_UNINITIALISED )
	&& ( state != STATE_ERROR )
	&& ( state != STATE_CONFIG_ERROR ))
	{
		(void) dvd_quit ( player_info ) ;

		pthread_join(read_thread, NULL); 
		pthread_join(write_thread, NULL);  

		output_deinitialise ( player_info -> out_plugin ) ;
	
		/* This isn't going to free everything.
		 * For instance the buffers will remain allocated.
		 */
		free ( player_info ) ;		

		state = STATE_UNINITIALISED ;
	}
}

static int get_information ( char * filename , int * major , int * minor )
{
	(* major) = (* minor) = 0 ;

	if ( strncmp ( player_info -> dvd_mountpoint , filename , 
					strlen ( player_info -> dvd_mountpoint ) ) == 0 )
	{
		/* it's on the DVD filesystem.. */

		gchar * name ;

		name = rindex ( filename , '/' ) ;
		if ( name != NULL )
		{
			if ( strncmp ( & name [ 1 ] , TITLE_IDENTIFIER , strlen ( TITLE_IDENTIFIER ) ) == 0 )
			{
				gchar * ext ;

				ext = rindex ( filename , ' ' ) ;
				if ( ext != NULL )
				{
					(* major) = atoi ( ext ) ;
					if (( (* major) < 1 )
					|| ( (* major) > 99 ))
					{
						(* major) = 0 ;
						return INFO_UNKNOWN ;
					} else {
						return INFO_TITLE ;
					}
				}
			} else if ( strncmp ( & name [ 1 ] , VTS_DIRECT_IDENTIFIER , strlen ( VTS_DIRECT_IDENTIFIER ) ) == 0 )
			{
				gchar * ext ;

				ext = rindex ( filename , ' ' ) ;
				if ( ext != NULL )
				{
					(* major) = atoi ( ext ) ;
					if (( (* major) < 0 )
					|| ( (* major) > 99 ))
					{
						(* major) = 0 ;
						return INFO_UNKNOWN ;
					} else {
						ext = rindex ( filename , '.' ) ;
						if ( ext != NULL )
						{
							(* minor) = atoi ( & ext [ 1 ] ) ;
							if (( (* minor) < 0 )
							|| ( (* minor) > 9 ))
							{
								(* minor) = 0 ;
								return INFO_UNKNOWN ;
							} else {
								return INFO_VTS_DIRECT ;
							}
						}
					}
				}
			}
		}
	}
	if ( buf_get_type_from_name ( filename ) != TYPE_UNKNOWN )
	{
		return INFO_VOB_DIRECT ;
	}
	return INFO_UNKNOWN ;
}

static int is_our_file(char *filename)
{
	int		t1 , t2 ;

	return ( get_information ( filename , & t1 , & t2 ) != INFO_UNKNOWN ) ;
}

static GList *scan_dir(char *dir)
{
	GList 		* list = NULL;
	gint		title ;
	char		* temp1 ;
	char		* temp2 ;
	gint		i ;

	if ( strlen ( player_info -> dvd_mountpoint ) < 1 )
	{
		return NULL ;
	}
	temp1 = strdup ( dir ) ;
	temp2 = strdup ( player_info -> dvd_mountpoint ) ;
	remove_trailing_slashes ( temp1 ) ;
	remove_trailing_slashes ( temp2 ) ;

	if (( strlen ( temp1 ) != strlen ( temp2 ) )
	|| ( strncmp ( temp1 , temp2 , strlen ( temp1 ) ) != 0 ))
	{
		free ( temp1 ) ;
		free ( temp2 ) ;
		return NULL;
	}

	free ( temp1 ) ;
	free ( temp2 ) ;
	/* Dump a list of titles present. */

	for ( title = 1 ; title < 100 ; title ++ )
	{
		int		title_fd ;
		ifo_t	* ifo_ptr ;

		title_fd = 0 ;
		ifo_ptr = ifo_examine_title ( player_info , title , & title_fd ) ;
		if ( ifo_ptr == NULL )
		{
			break ;
		}
		list = g_list_append(list, g_strdup_printf( "%s%d" , TITLE_IDENTIFIER , title )) ;
		ifoClose ( ifo_ptr ) ;
		close ( title_fd ) ;
	}

	/* Dump a list of VTS files present */
	temp1 = malloc ( strlen ( player_info -> dvd_mountpoint ) + 34 ) ;
	temp2 = malloc ( 32 ) ;
	for ( title = 0 ; title < 100 ; title ++ )
	{
		for ( i = 0 ; ( title > 0 ) ? ( i < 10 ) : ( i < 1 ) ; i ++ )
		{
			struct stat	statbuf ;

			if ( title == 0 )
			{
				snprintf ( temp2 , 32 , VIDEO_TS_NAME ) ;
			} else {
				snprintf ( temp2 , 32 , VTS_NAME , title , i ) ;
			}
			strcpy ( temp1 , player_info -> dvd_mountpoint ) ;
			strcat ( temp1 , temp2 ) ;
			if ( stat ( temp1 , & statbuf ) == 0 )
			{
				list = g_list_append(list, g_strdup_printf( "%s%d.%d" , VTS_DIRECT_IDENTIFIER , title , i )) ;
			} else {
				string_uppercase ( temp2 ) ;
				strcpy ( temp1 , player_info -> dvd_mountpoint ) ;
				strcat ( temp1 , temp2 ) ;
				if ( stat ( temp1 , & statbuf ) == 0 )
				{
					list = g_list_append(list, g_strdup_printf( "%s%d.%d" , VTS_DIRECT_IDENTIFIER , title , i )) ;
				} else {
					break ;
				}
			}
		}
		if ( i == 0 )
		{
			/* nothing read in this title. exit */
			break ;
		}
	}
	free ( temp1 ) ;
	free ( temp2 ) ;

	return list ;
}

static void play_file(char *filename)
{
	char		errbuf [ BUFFER_LENGTH ] ;
	int			title , minor ;
	char		* flist [ 1 ] ;
	char		* t = NULL ;

	if ( ! is_our_file ( filename ) )
	{
		return ;
	}

	minidxr2_deferred_init () ;

	if ( state != STATE_READY )
	{
		return ;
	}

	/* Try to identify this file as a DVD title first */
	switch ( get_information ( filename , & title , & minor ) )
	{
		case INFO_TITLE :
					if ( ! dvd_open_title ( player_info , title , errbuf , BUFFER_LENGTH_1 ) )
					{
						static GtkWidget * dialog;

						if (dialog != NULL)
							return;

						dialog = xmms_show_message ( "Play Title Error" ,
									errbuf , "Ok" , FALSE , NULL , NULL ) ;

						gtk_signal_connect ( GTK_OBJECT ( dialog ) , "destroy" ,
								   GTK_SIGNAL_FUNC ( gtk_widget_destroyed ) ,
								   & dialog ) ;
						return ;
					}
					break ;
		case INFO_VOB_DIRECT :
					/* not a title or a video - must be able to play it directly. */
					flist [ 0 ] = filename ;
					if ( ! dvd_open_file ( player_info , flist , 1 , errbuf , BUFFER_LENGTH_1 ) )
					{
						static GtkWidget * dialog;

						if (dialog != NULL)
							return;

						dialog = xmms_show_message ( "Play File Error" ,
									errbuf , "Ok" , FALSE , NULL , NULL ) ;

						gtk_signal_connect ( GTK_OBJECT ( dialog ) , "destroy" ,
								   GTK_SIGNAL_FUNC ( gtk_widget_destroyed ) ,
								   & dialog ) ;
						return ;
					}
					break ;
		case INFO_VTS_DIRECT :
					{
						struct stat	statbuf ;
						char * temp1 ;
						char * temp2 ;

						temp1 = malloc ( strlen ( player_info -> dvd_mountpoint ) + 34 ) ;
						temp2 = malloc ( 32 ) ;
						if ( title == 0 )
						{
							snprintf ( temp2 , 32 , VIDEO_TS_NAME ) ;
						} else {
							snprintf ( temp2 , 32 , VTS_NAME , title , minor ) ;
						}
						strcpy ( temp1 , player_info -> dvd_mountpoint ) ;
						strcat ( temp1 , temp2 ) ;
						if ( stat ( temp1 , & statbuf ) == 0 )
						{
							flist [ 0 ] = temp1 ;
						} else {
							string_uppercase ( temp2 ) ;
							strcpy ( temp1 , player_info -> dvd_mountpoint ) ;
							strcat ( temp1 , temp2 ) ;
							flist [ 0 ] = temp1 ;
						}
						free ( temp2 ) ;
						if ( ! dvd_open_file ( player_info , flist , 1 , errbuf , BUFFER_LENGTH_1 ) )
						{
							static GtkWidget * dialog;

							free ( temp1 ) ;
							if (dialog != NULL)
								return;

							dialog = xmms_show_message ( "Play VTS Error" ,
										errbuf , "Ok" , FALSE , NULL , NULL ) ;

							gtk_signal_connect ( GTK_OBJECT ( dialog ) , "destroy" ,
									   GTK_SIGNAL_FUNC ( gtk_widget_destroyed ) ,
									   & dialog ) ;
							return ;
						}

						free ( temp1 ) ;
					}
					break ;
	}
	dvd_play ( player_info ) ;

	get_song_info ( filename , & t , & current_time_length ) ;


	{
		char * str = NULL ;

		parse_reverse ( ifo_settings_audio_freq ( player_info , 
						player_info -> current_title , current_audio_track ) ,
					audio_freq_table , & str , NULL , NULL ) ;
		if ( str == NULL )
		{
			str = "48" ;
		}
		current_audio_freq = atoi ( str ) * 1000 ;
	}
}

static void stop(void)
{
	if ( state != STATE_READY )
	{
		return ;
	}

	dvd_unload ( player_info ) ;
}

static void minidxr2_pause(short p)
{
	if ( state != STATE_READY )
	{
		return ;
	}

	dvd_pause ( player_info ) ;
}

static void seek(int time)
{
	if ( ifo_available ( player_info ) )
	{
		int		hour , min , sec ;

		if ( time < 0 )
		{
			return ;
		}
		sec = ( time % 60 ) ;
		min = (( time / 60 ) % 60 ) ;
		hour = ( time / 3600 ) ;

		dvd_seek_time ( player_info , hour , min , sec ) ;
	} else {
		int			pos , length = 0 ;

		if (( dvd_get_length ( player_info , & length ) == NULL )
		&& ( length > 0 ))
		{
			pos = ((( time * DEFAULT_SUB_LENGTH ) / ( DEFAULT_LENGTH / 1000 ))
					* length ) / DEFAULT_SUB_LENGTH ;
			if (( pos >= 0 )
			&& ( pos < length ))
			{
				dvd_seek_position ( player_info , pos ) ;
			}
		}
	}
}


static int get_time(void)
{
	int			pos = 0 ;
	char		* ret ;
	int			rc , length = 0 ;
	int			hour , min , sec ;

	ret = dvd_get_position ( player_info , & pos ) ;
	if ( ret != NULL )
	{
		return ( 0 ) ;
	}

	if ( ifo_available ( player_info ) )
	{
		minidxr2_ip . set_info ( NULL , current_time_length , 
				output_get_bitrate ( player_info -> out_plugin ) ,
				current_audio_freq , 2 ) ;

		rc = ifo_get_time_for_position ( player_info ,
				player_info -> current_title , pos , & hour , & min , & sec ) ;
		if ( rc == IFO_OK )
		{
			return (( sec + ( min * 60 ) + ( hour * 3600 )) * 1000 ) ;
		}
	} else {
		if (( dvd_get_length ( player_info , & length ) != NULL )
		&& ( length > 0 ))
		{
			minidxr2_ip . set_info ( NULL , current_time_length , 
					output_get_bitrate ( player_info -> out_plugin ) ,
					48000 , 2 ) ;

			rc = ((( pos * DEFAULT_SUB_LENGTH ) / length ) 
						* ( DEFAULT_LENGTH / DEFAULT_SUB_LENGTH )) ;

			return rc ;
		}
	}
	return 0 ;
}

static void get_song_info(char *filename, char **title, int *len)
{
	int			title_num , minor ;

	(* title) = NULL ;
	switch ( get_information ( filename , & title_num , & minor ) )
	{
		case INFO_TITLE :
					{
						int		title_num_fd ;
						ifo_t	* ifo_ptr ;
						int		sec , min , hour ;

						printf ( "Title decoding %d\n" , title_num ) ;
						title_num_fd = 0 ;
						ifo_ptr = ifo_examine_title ( player_info , title_num , & title_num_fd ) ;
						if ( ifo_ptr == NULL )
						{
							return ;
						}
						if ( ifo_direct_get_time_for_position ( ifo_ptr , -1 , & hour , & min , & sec ) 
									== IFO_OK )
						{
							(* len) = ( sec + ( min * 60 ) + ( hour * 3600 )) * 1000 ;
						}
						ifoClose ( ifo_ptr ) ;
						close ( title_num_fd ) ;
					}
					break ;
		case INFO_VTS_DIRECT :
		case INFO_VOB_DIRECT :
					{
						(* len) = DEFAULT_LENGTH ;
					}
					break ;
	}
}

/* static void get_volume(int *l, int *r)
{
	int fd, devs, cmd, v;

#if defined(HAVE_SYS_SOUNDCARD_H) || defined(HAVE_MACHINE_SOUNDCARD_H)
	if (minidxr2_cfg.use_oss_mixer)
	{
		fd = open(DEV_MIXER, O_RDONLY);
		if (fd != -1)
		{
			ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
			if (devs & SOUND_MASK_CD)
				cmd = SOUND_MIXER_READ_CD;
			else if (devs & SOUND_MASK_VOLUME)
				cmd = SOUND_MIXER_READ_VOLUME;
			else
			{
				close(fd);
				return;
			}
			ioctl(fd, cmd, &v);
			*r = (v & 0xFF00) >> 8;
			*l = (v & 0x00FF);
			close(fd);
		}
	}
	else
#endif
	if (!minidxr2_cfg.use_oss_mixer)
	{

#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
		struct ioc_vol vol;

		if (minidxr2_fd != -1)
		{
			ioctl(minidxr2_fd, CDIOCGETVOL, &vol);
			*l = (100 * vol.vol[0]) / 255;
			*r = (100 * vol.vol[1]) / 255;
		}
#elif defined(CDROMVOLREAD)
		struct cdrom_volctrl vol;

		if (minidxr2_fd != -1)
		{
			ioctl(minidxr2_fd, CDROMVOLREAD, &vol);
			*l = (100 * vol.channel0) / 255;
			*r = (100 * vol.channel1) / 255;
		}
#else
		*l = volume_left;
		*r = volume_right;
#endif

	}
}

static void set_volume(int l, int r)
{
#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
	struct ioc_vol vol;
#else
	struct cdrom_volctrl vol;
#endif
	int fd, devs, cmd, v;

#if defined(HAVE_SYS_SOUNDCARD_H) || defined(HAVE_MACHINE_SOUNDCARD_H)
	if (minidxr2_cfg.use_oss_mixer)
	{
		fd = open(DEV_MIXER, O_RDONLY);
		if (fd != -1)
		{
			ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
			if (devs & SOUND_MASK_CD)
				cmd = SOUND_MIXER_WRITE_CD;
			else if (devs & SOUND_MASK_VOLUME)
				cmd = SOUND_MIXER_WRITE_VOLUME;
			else
			{
				close(fd);
				return;
			}
			v = (r << 8) | l;
			ioctl(fd, cmd, &v);
			close(fd);
		}
	}
	else
#endif
	{
		if (minidxr2_fd != -1)
		{
#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
			vol.vol[0] = vol.vol[2] = (l * 255) / 100;
			vol.vol[1] = vol.vol[3] = (r * 255) / 100;
			ioctl(minidxr2_fd, CDIOCSETVOL, &vol);
#else
			vol.channel0 = vol.channel2 = (l * 255) / 100;
			vol.channel1 = vol.channel3 = (r * 255) / 100;
			ioctl(minidxr2_fd, CDROMVOLCTRL, &vol);
#endif
		}
#if !defined(CDROMVOLREAD) && !(defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__))
		volume_left = l;
		volume_right = r;
#endif
	}
} */

